/*************************************************************************
 * DobHelper
 * Copyright 2013 Anthony Amundson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version
 ************************************************************************/

 package dobhelper;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.image.BufferedImage;
import java.awt.print.Book;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.LinkedList;
import java.util.TimeZone;
import javax.imageio.ImageIO;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.SpinnerDateModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumnModel;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import stellarCartography.Constellation;
import stellarCartography.EntityLibrary;
import stellarCartography.ObservationList;
import stellarCartography.ObservationPoint;
import stellarCartography.Star;
import stellarCartography.StellarEntity;
import util.DebugLogger;

/* TODO: closing the map window does not 'uncheck' the map window shown boxes */


public class DobHelper extends JFrame implements ActionListener, 
		ChangeListener, WindowListener, MapSettingsChangedListener, 
		TreeSelectionListener, MouseListener,
		StellarEntitySelectionListener, MapSelectionChangeListener{

	public static final String userDirName = System.getProperty("user.home") + File.separator + ".DobHelp";

	/* GUI data */
	private JMenuBar menuBar;
	private JMenu fileMenu;
	private JMenuItem exportMapMenuItem;
	private JMenuItem printMapMenuItem;
	private JMenuItem exitMenuItem;
	private JMenu mapMenu;
	private JCheckBoxMenuItem showMapWindMenuItem;
	private JMenuItem addSelectionMenuItem;
	private JMenuItem removeSelectionMenuItem;
	private JMenuItem removeAllEntriesMenuItem;
	private JMenuItem mapSettingsMenuItem;
	private JMenu listMenu;
	private JMenuItem addListMenuItem;
	private JMenuItem removeListMenuItem;
	private JMenu conditionsMenu;
	private JMenuItem timezoneMenuItem;
	private JMenu helpMenu;
	private JMenuItem aboutMenuItem;

	
	/* control panel contents */
	private JLabel latLbl;
	private JSpinner latSpnr;
	private JLabel lonLbl;
	private JSpinner lonSpnr;
	private JLabel yearLbl;
	private JSpinner yearSpnr;
	private JButton clearSelectionBtn;
	private JCheckBox showMapBox;

	private JTabbedPane tabController;

	/* map stuff */
	private MapWindow mapWindow;
	
	/* details stuff */
	private JPanel detailPanel;
	private JTextField detailName;
	private JSpinner detailRank;
	private JList detailDesignations;
	private JCheckBox detailSeen;
	private JTextArea detailNotes;
	
	private JLabel detailNameLbl;
	private JLabel detailRankLbl;
	private JPanel detailDesignationsPanel;
	private JPanel detailNotesPanel;
	
	private JButton detailCommit;
	
	/* list browser */
	private JTree listTree;
	private JScrollPane listTreeScroller;
	private JPanel listTreePanel;
	private JSplitPane horizontalSplit;
	private JPopupMenu tablePopup;
	private JMenuItem addUserListItem;
	private JMenuItem removeUserListItem;

	/* non-GUI stuff */
	private EntityLibrary library;
	private MapSettings mapSettings;
	private double timezoneOffset;
	private boolean isDST;
	TableSetManager tableSetManager;


	public DobHelper() {
		super();

		DebugLogger.setLoggingEnabled(true);

		library = new EntityLibrary(userDirName);
		Constellation.initConstellations(library);

		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				setLayout(new BorderLayout());
				setTitle(Main.PROG_TEXT);

				/* setup the menus */
				createMenus();
				setJMenuBar(menuBar);

				JToolBar obsCoordToolbar = new JToolBar("Coordinates:");
				//controlPanel = new JPanel();
				obsCoordToolbar.setLayout(new FlowLayout());
				latLbl = new JLabel("Latitude");
				obsCoordToolbar.add(latLbl);
				latSpnr = new JSpinner(new SpinnerNumberModel(45.0, -90.0, 90.0, 1.0));
				latSpnr.addChangeListener(DobHelper.this);
				obsCoordToolbar.add(latSpnr);
				lonLbl = new JLabel("Longitude");
				obsCoordToolbar.add(lonLbl);
				lonSpnr = new JSpinner(new SpinnerNumberModel(-93.5, -180.0, 180.0, 1.0));
				lonSpnr.addChangeListener(DobHelper.this);
				obsCoordToolbar.add(lonSpnr);
				yearLbl = new JLabel("Time");
				obsCoordToolbar.add(yearLbl);
				yearSpnr = new JSpinner(new SpinnerDateModel(new Date(),
						new GregorianCalendar(1900, 1, 1).getTime(),
						new GregorianCalendar(2100, 1, 1).getTime(),
						Calendar.DAY_OF_YEAR));
				yearSpnr.addChangeListener(DobHelper.this);
				obsCoordToolbar.add(yearSpnr);

				obsCoordToolbar.addSeparator();

				//JToolBar mapToolbar = new JToolBar("Map Settings");
				clearSelectionBtn = new JButton("Clear Map");
				clearSelectionBtn.addActionListener(DobHelper.this);
				obsCoordToolbar.add(clearSelectionBtn);
				showMapBox = new JCheckBox("Show Map Window");
				showMapBox.setSelected(false);
				showMapBox.addActionListener(DobHelper.this);
				obsCoordToolbar.add(showMapBox);

				//add(controlPanel, BorderLayout.NORTH);
				obsCoordToolbar.setFloatable(false);
				add(obsCoordToolbar, BorderLayout.PAGE_START);
				//add(mapToolbar, BorderLayout.PAGE_START);

				TimeZone tz = Calendar.getInstance().getTimeZone();
				timezoneOffset = tz.getRawOffset() / 3600000; /* convert from ms to hours */
				isDST = (tz.getDSTSavings() > 0);

				mapSettings = new MapSettings();
				mapSettings.updateLocation(getCurrentObsPoint());

				tabController = new JTabbedPane();

				tableSetManager = new TableSetManager(tabController, mapSettings);
				tableSetManager.addSelectionListener(DobHelper.this);
				tableSetManager.addMapSelectionChangeListener(DobHelper.this);
				
				listTreePanel = new JPanel();
				listTreePanel.setPreferredSize(new Dimension(200, 1000));
				listTreePanel.setLayout(new BorderLayout());

				listTree = new JTree(generateListTree());
				listTree.addMouseListener(DobHelper.this);
				listTreeScroller = new JScrollPane(listTree);
				listTreePanel.add(listTreeScroller, BorderLayout.CENTER);
				listTree.setTransferHandler(new ObsListTreeTransferHandler(listTree, library));
				
				tablePopup = new JPopupMenu();
				addUserListItem = new JMenuItem("Add Observation List");
				addUserListItem.addActionListener(DobHelper.this);
				tablePopup.add(addUserListItem);
				removeUserListItem = new JMenuItem("Remove Observation List");
				removeUserListItem.addActionListener(DobHelper.this);
				removeUserListItem.setEnabled(false);
				tablePopup.add(removeUserListItem);
				listTree.setComponentPopupMenu(tablePopup);
				
				listTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
				listTree.addTreeSelectionListener(DobHelper.this);
				
				horizontalSplit = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, listTreePanel, tabController);
				add(horizontalSplit, BorderLayout.CENTER);
				
				detailPanel = new JPanel();
				JPanel tempPanel = new JPanel(null);
				detailName = new JTextField("");
				detailName.setBounds(110, 5, 145, 20);
				detailName.setEditable(false);
				tempPanel.add(detailName);
				detailRank = new JSpinner(new SpinnerNumberModel(10, 1, 10, 1));
				detailRank.setBounds(110, 30, 145, 20);
				tempPanel.add(detailRank);
				detailDesignations = new JList();
				detailDesignations.setBounds(10, 20, 230, 100);
				detailSeen = new JCheckBox("Object has been observed");
				detailSeen.setBounds(260, 5, 200, 20);
				tempPanel.add(detailSeen);
				detailCommit = new JButton("Save Observation Changes");
				detailCommit.addActionListener(DobHelper.this);
				detailCommit.setBounds(465, 5, 195, 20);
				tempPanel.add(detailCommit);
				detailNotes = new JTextArea("");
				detailNotes.setAlignmentY(TOP_ALIGNMENT);
				JScrollPane notesScroller = new JScrollPane(detailNotes);
				notesScroller.setBounds(10, 20, 380, 125);
				
				detailNameLbl = new JLabel("Name: ");
				detailNameLbl.setBounds(5, 5, 100, 20);
				tempPanel.add(detailNameLbl);
				detailRankLbl = new JLabel("Rank: ");
				detailRankLbl.setBounds(5, 30, 100, 20);
				tempPanel.add(detailRankLbl);
				detailDesignationsPanel = new JPanel(null);
				detailDesignationsPanel.setBorder(new TitledBorder("Other Designations"));
				detailDesignationsPanel.add(detailDesignations);
				detailDesignationsPanel.setBounds(5, 55, 250, 130);
				tempPanel.add(detailDesignationsPanel);
				detailNotesPanel = new JPanel(null);
				detailNotesPanel.setBorder(new TitledBorder("Observation Notes"));
				detailNotesPanel.setBounds(260, 30, 400, 155);
				detailNotesPanel.add(notesScroller);
				tempPanel.add(detailNotesPanel);

				tempPanel.setPreferredSize(new Dimension(670, 195));
				detailPanel.add(tempPanel);

				add(detailPanel, BorderLayout.SOUTH);

				updateDetailsPane();

				setSize(new Dimension(1024, 768));
				setResizable(true);
				setIconImage(Toolkit.getDefaultToolkit().getImage("icons" + File.separator + "DobHelper.png"));

				setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
				setVisible(true);
			}
		});
	}
	
	private void updateDetailsPane(){
		StellarEntity entity = tableSetManager.getSelectedEntry();
		if (entity != null){
			detailName.setText(entity.getName());
			double distance = entity.getDistance();
			String[] designations = entity.getDesignations();
			detailDesignations.setListData(designations);
			detailSeen.setSelected(entity.getIsObserved());
			detailRank.setValue(entity.getRank());
			detailNotes.setText(entity.getNotes());
			return;
		}
		
		
		detailName.setText("Select an Entry");
		detailDesignations.setListData(new String[0]);
		detailSeen.setSelected(false);
		detailNotes.setText("");
		return;
	}
	
	private void saveDetails(){
		StellarEntity entity = tableSetManager.getSelectedEntry();
		if (entity == null){
			return;
		}

		/* update library and tables */
		entity.setObserved(detailSeen.isSelected());
		entity.setNotes(detailNotes.getText());
		entity.setRank(((Integer)detailRank.getValue()).intValue());

		tableSetManager.updateSelectedRowContents();
			
		/* save the contents to file as well */
		String userDirName = System.getProperty("user.home") + File.separator + ".DobHelp";
		File userDir = new File(userDirName);
		if (!userDir.exists()){
			/* try creating the directory */
			if (!userDir.mkdirs()){
				DebugLogger.logTextLine("Error creating directory " + userDirName + ".");
				return;
			}

		} else if (!userDir.isDirectory()){
			/* path already exists, and is dir!!! */
			DebugLogger.logTextLine("Cannot save user detail entries " + userDirName + " already exists and is not a directory.");
			return;
		}

		/* at this point dir should exist - look for existing notes file */
		String noteFileName = userDirName + File.separator + EntityLibrary.NOTES_FILENAME;
		File noteFile = new File(noteFileName);
		if (!noteFile.exists()){
			/* file does not exist -- create it with this note as the first entry */
			library.saveNotesFirstDetailEntry(noteFile, entity);
		} else {
			library.updateNotesEntry(noteFile, entity);
		}
	}

	private void createMenus(){
		menuBar = new JMenuBar();

		fileMenu = new JMenu("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		menuBar.add(fileMenu);
		
		exportMapMenuItem = new JMenuItem("Export Map As...");
		exportMapMenuItem.addActionListener(this);
		exportMapMenuItem.setMnemonic(KeyEvent.VK_E);
		exportMapMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_E, ActionEvent.CTRL_MASK));
		fileMenu.add(exportMapMenuItem);

		printMapMenuItem = new JMenuItem("Print Map");
		printMapMenuItem.addActionListener(this);
		printMapMenuItem.setMnemonic(KeyEvent.VK_P);
		printMapMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P, ActionEvent.CTRL_MASK));
		fileMenu.add(printMapMenuItem);

		fileMenu.addSeparator();

		exitMenuItem = new JMenuItem("Quit");
		exitMenuItem.addActionListener(this);
		exitMenuItem.setMnemonic(KeyEvent.VK_Q);
		exitMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
		fileMenu.add(exitMenuItem);

		mapMenu = new JMenu("Map");
		mapMenu.setMnemonic(KeyEvent.VK_M);
		menuBar.add(mapMenu);

		showMapWindMenuItem = new JCheckBoxMenuItem("Show Map Window");
		showMapWindMenuItem.addActionListener(this);
		showMapWindMenuItem.setMnemonic(KeyEvent.VK_S);
		showMapWindMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.CTRL_MASK));
		mapMenu.add(showMapWindMenuItem);
		mapMenu.addSeparator();
		addSelectionMenuItem = new JMenuItem("Add Selection to Map");
		addSelectionMenuItem.addActionListener(this);
		addSelectionMenuItem.setMnemonic(KeyEvent.VK_A);
		addSelectionMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T, ActionEvent.CTRL_MASK));
		mapMenu.add(addSelectionMenuItem);
		removeSelectionMenuItem = new JMenuItem("Remove Selection from Map");
		removeSelectionMenuItem.addActionListener(this);
		removeSelectionMenuItem.setMnemonic(KeyEvent.VK_R);
		removeSelectionMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, ActionEvent.CTRL_MASK));
		mapMenu.add(removeSelectionMenuItem);
		removeAllEntriesMenuItem = new JMenuItem("Remove All Entries from Map");
		removeAllEntriesMenuItem.addActionListener(this);
		removeAllEntriesMenuItem.setMnemonic(KeyEvent.VK_N);
		removeAllEntriesMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, ActionEvent.CTRL_MASK));
		mapMenu.add(removeAllEntriesMenuItem);
		mapMenu.addSeparator();
		mapSettingsMenuItem = new JMenuItem("Configure Map Settings");
		mapSettingsMenuItem.addActionListener(this);
		mapSettingsMenuItem.setMnemonic(KeyEvent.VK_C);
		mapMenu.add(mapSettingsMenuItem);


		listMenu = new JMenu("Lists");
		listMenu.setMnemonic(KeyEvent.VK_L);
		menuBar.add(listMenu);
		addListMenuItem = new JMenuItem("Add User Observation List");
		addListMenuItem.addActionListener(this);
		addListMenuItem.setMnemonic(KeyEvent.VK_A);
		listMenu.add(addListMenuItem);
		removeListMenuItem = new JMenuItem("Remove User Observation List");
		removeListMenuItem.addActionListener(this);
		removeListMenuItem.setMnemonic(KeyEvent.VK_R);
		removeListMenuItem.setEnabled(false);
		listMenu.add(removeListMenuItem);
		
		conditionsMenu = new JMenu("Conditions");
		conditionsMenu.setMnemonic(KeyEvent.VK_O);
		menuBar.add(conditionsMenu);
		
		timezoneMenuItem = new JMenuItem("Set Timezone");
		timezoneMenuItem.addActionListener(this);
		timezoneMenuItem.setMnemonic(KeyEvent.VK_T);
		conditionsMenu.add(timezoneMenuItem);
		
		helpMenu = new JMenu("Help");
		helpMenu.setMnemonic(KeyEvent.VK_H);
		menuBar.add(helpMenu);
		
		aboutMenuItem = new JMenuItem("About");
		aboutMenuItem.addActionListener(this);
		aboutMenuItem.setMnemonic(KeyEvent.VK_A);
		helpMenu.add(aboutMenuItem);
	}
	
	private void exportMap(){
		JFileChooser chooser = new JFileChooser();
		chooser.setMultiSelectionEnabled(false);
		chooser.setAcceptAllFileFilterUsed(true);
		FileNameExtensionFilter filter = new FileNameExtensionFilter("PNG images", "png");
		chooser.setFileFilter(filter);
		int returnVal = chooser.showOpenDialog(this);
		
		if (returnVal == JFileChooser.APPROVE_OPTION){
			File file;
			String fileName = chooser.getSelectedFile().getName();
			if (chooser.getFileFilter() == filter){
				int periodIndex = fileName.lastIndexOf('.');
				if (periodIndex < 0 || !fileName.substring(periodIndex).equals(".png")){
					file = new File(chooser.getSelectedFile() + ".png");
				} else {
					file = chooser.getSelectedFile();
				}
			} else {
				file = chooser.getSelectedFile();
			}
			
			/* TODO: proper way to get size... */
			BufferedImage image = new BufferedImage(712, 712, BufferedImage.TYPE_INT_ARGB);
			
			Graphics2D g2d = image.createGraphics();
			
			MapPlotter.drawMap(g2d, 712, 712, mapSettings);
			
			try{
				ImageIO.write(image, "png", file);
			} catch (Exception e){
				DebugLogger.logText(e.getMessage());
			}
		}
	}

	private void printMap(){
		PrinterJob job = PrinterJob.getPrinterJob();
		Book book = new Book();
		PageFormat pageFormat = job.defaultPage();
		Paper paper = pageFormat.getPaper();
		paper.setImageableArea(36, 36, pageFormat.getWidth() - 72, pageFormat.getHeight() - 72);
		pageFormat.setPaper(paper);
		book.append(new MapPrinter(mapSettings), pageFormat, 1);
		job.setPageable(book);
		boolean doPrint = job.printDialog();
		if (doPrint){
			try {
				job.print();
			} catch (PrinterException pe){
				DebugLogger.logText(pe.getMessage());
			}
		}
	}

	private void quit(){
		System.exit(0);
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == exportMapMenuItem){
			exportMap();
		} else if (e.getSource() == printMapMenuItem){
			printMap();
		} else if (e.getSource() == exitMenuItem){
			quit();
		} else if (e.getSource() == showMapWindMenuItem){
			if (((JCheckBoxMenuItem)e.getSource()).isSelected()){
				mapWindow = new MapWindow(getMapSettings());
				mapWindow.addWindowListener(this);
				mapWindow.setVisible(true);
			} else {
				if (mapWindow != null){
					mapWindow.setVisible(false);
					mapWindow.dispose();
				}
			}
			showMapBox.setSelected(showMapWindMenuItem.isSelected());
		} else if (e.getSource() == showMapBox){
			if (showMapBox.isSelected()){
				mapWindow = new MapWindow(getMapSettings());
				mapWindow.addWindowListener(this);
				mapWindow.setVisible(true);

			} else {
				if (mapWindow != null){
					mapWindow.setVisible(false);
					mapWindow.dispose();
				}
			}
			showMapWindMenuItem.setSelected(showMapBox.isSelected());
		} else if (e.getSource() == mapSettingsMenuItem){
			MapSettingsDialog dialog = new MapSettingsDialog(getMapSettings());
			dialog.addChangeListener(this);
			dialog.setVisible(true);
		} else if (e.getSource() == detailCommit){
			saveDetails();
		} else if (e.getSource() == timezoneMenuItem){
			TimeZoneDialog tzd = new TimeZoneDialog(this, timezoneOffset, isDST);
			tzd.setVisible(true);
			if (tzd.isAccepted()){
				timezoneOffset = tzd.getOffset();
				isDST = tzd.getDST();

				ObservationPoint point = getCurrentObsPoint();
				tableSetManager.updatePoint(point);
				mapSettings.updateLocation(point);
				updateMaps();
			} 
		} else if (e.getSource() == aboutMenuItem){
			AboutDialog helpD = new AboutDialog(this);
			helpD.setVisible(true);
		} else if (e.getSource() == addSelectionMenuItem){
			tableSetManager.addSelectedItems();
		} else if (e.getSource() == removeSelectionMenuItem){
			tableSetManager.removeSelectedItems();
		} else if (e.getSource() == clearSelectionBtn || e.getSource() == removeAllEntriesMenuItem){
			tableSetManager.clearAllSelections();
		} else if (e.getSource() == addUserListItem || e.getSource() == addListMenuItem){
			addEmptyUserList();
		} else if (e.getSource() == removeUserListItem || e.getSource() == removeListMenuItem){
			removeSelectedUserList();
		}
	}

	public void stateChanged(ChangeEvent e) {
		if (e.getSource() == latSpnr || e.getSource() == lonSpnr ||
				e.getSource() == yearSpnr){
			ObservationPoint point = getCurrentObsPoint();
			tableSetManager.updatePoint(point);
			mapSettings.updateLocation(point);
			updateMaps();
		}
	}

	public ObservationPoint getCurrentObsPoint(){
		return new ObservationPoint(((Double)latSpnr.getValue()).doubleValue(),
						((Double)lonSpnr.getValue()).doubleValue(),
						(Date)yearSpnr.getValue(), timezoneOffset, isDST);
	}

	public void windowOpened(WindowEvent e) {
		if (e.getSource() == mapWindow){
			showMapWindMenuItem.setSelected(true);
			moveWindowToOtherScreen(mapWindow);
		}
	}

	public void windowClosing(WindowEvent e) {
		if (e.getSource() == mapWindow){
			showMapWindMenuItem.setSelected(false);
		}
	}

	public void windowClosed(WindowEvent e) {
		if (e.getSource() == mapWindow){
			showMapWindMenuItem.setSelected(false);
		}
	}

	public void windowIconified(WindowEvent e) {
	}

	public void windowDeiconified(WindowEvent e) {
	}

	public void windowActivated(WindowEvent e) {
	}

	public void windowDeactivated(WindowEvent e) {
	}

	private void moveWindowToOtherScreen(JFrame frame){
		int deviceNum = -1;

		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice[] devices = ge.getScreenDevices();
		if (devices.length < 2){
			/* there is only one screen */
			return;
		}

		/* look through all screens and configs to find the one currently in use */
		for (int j = 0 ; j < devices.length ; j++){
			GraphicsDevice device = devices[j];
			GraphicsConfiguration[] gc = device.getConfigurations();
			for (int i = 0 ; i < gc.length ; i++){
				Rectangle bounds = gc[i].getBounds();
				if (bounds.contains(this.getLocation())){
					deviceNum = j;
				}
			}
		}

		/* move the window to the lowest number screen not in use */
		deviceNum = (deviceNum == 0 ? 1 : 0);
		Rectangle rect = devices[deviceNum].getConfigurations()[0].getBounds();
		mapWindow.setLocation(rect.getLocation());
	}

	private void updateMaps(){
		MapSettings mapSettings = getMapSettings();
		if (mapWindow != null && mapWindow.isVisible()){
			mapWindow.updateContents(mapSettings);
		}
		tableSetManager.updateMapSettings(mapSettings);
	}

	private MapSettings getMapSettings(){
		LinkedList<StellarEntity> entities = tableSetManager.getVisibleEntities();

		mapSettings.setDisplayEntities(entities);
		mapSettings.setDisplayStars(createStarList());

		return mapSettings;
	}

	private LinkedList<Star> createStarList(){
		LinkedList<Star> retList = library.getStars(mapSettings.getMinimumMag());
		
		if (mapSettings.getShowConstellationStars()){
			Constellation.addMissingStars(retList, mapSettings.getMinimumMag());
		}

		return retList;
	}

	public void mapSettingsChanged(MapSettings settings) {
		updateMaps();
	}
	
	private DefaultMutableTreeNode generateListTree(){
		DefaultMutableTreeNode listTreeRoot = new DefaultMutableTreeNode("Library Contents");
		
		DefaultMutableTreeNode listsByType = new DefaultMutableTreeNode("By Type");
		listsByType.setAllowsChildren(true);
		
		DefaultMutableTreeNode listsByCatalog = new DefaultMutableTreeNode("By Catalog");
		listsByCatalog.setAllowsChildren(true);
		
		DefaultMutableTreeNode listsByUser = new DefaultMutableTreeNode("User Lists");
		listsByUser.setAllowsChildren(true);
		
		for (ObservationList list : library.getObLists()){
			ObservationListTreeNode newNode = new ObservationListTreeNode(list);
			switch(list.getListType()){
				case ObservationList.OBLIST_TYPE_BY_TYPE:
					listsByType.add(newNode);
					break;
				case ObservationList.OBLIST_TYPE_BY_CATALOG:
					listsByCatalog.add(newNode);
					break;
				case ObservationList.OBLIST_TYPE_USER:
					listsByUser.add(newNode);
					break;
				case ObservationList.OBLIST_TYPE_COMPLETE:
					listTreeRoot.add(newNode);
			}
		}
		
		listTreeRoot.add(listsByType);
		listTreeRoot.add(listsByCatalog);
		listTreeRoot.add(listsByUser);
		
		return listTreeRoot;
	}
	
	private void addEmptyUserList(){
		String name = JOptionPane.showInputDialog(this, "Please Enter List Name", 
				"Create New Observation List", JOptionPane.PLAIN_MESSAGE);
		if (name == null){
			/* user cancelled */
			return;
		}
		if (name.equals("")){
			JOptionPane.showMessageDialog(this, "ERROR: List names must have at least one charecter.", 
					"Error Creating Observation List", JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		/* validate that name is not a duplicate */
		if (library.containsUserListNamed(name)){
			JOptionPane.showMessageDialog(this, "ERROR: A list with that name already exists.", 
					"Error Creating Observation List", JOptionPane.ERROR_MESSAGE);
			return;
		}
		ObservationList newList = library.createNewObservationList(name, userDirName);
		ObservationListTreeNode newNode = new ObservationListTreeNode(newList);
		DefaultTreeModel model = (DefaultTreeModel)listTree.getModel();
		DefaultMutableTreeNode root = (DefaultMutableTreeNode)(model.getRoot());
		DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode)model.getChild(root, 3);
		model.insertNodeInto(newNode, parentNode, parentNode.getChildCount());
	}
	
	private void removeSelectedUserList(){
		DefaultMutableTreeNode selectedNode = (DefaultMutableTreeNode)listTree.getLastSelectedPathComponent();
		if (selectedNode != null && selectedNode instanceof ObservationListTreeNode){
			ObservationListTreeNode listNode = (ObservationListTreeNode)selectedNode;
			if (listNode.getObservationList().getListType() == ObservationList.OBLIST_TYPE_USER){
				/* confirm */
				int option = JOptionPane.showConfirmDialog(this, "Permenantly remove list: \"" +
						listNode.getObservationList().getListName() + "\"", "Really delete list?", JOptionPane.YES_NO_OPTION);

				if (option == JOptionPane.YES_OPTION){
					library.removeObservationList(listNode.getObservationList(), userDirName);
					DefaultTreeModel model = (DefaultTreeModel)listTree.getModel();
					model.removeNodeFromParent(listNode);
					tableSetManager.removeTabForList(listNode.getObservationList());
					removeUserListItem.setEnabled(false);
					removeListMenuItem.setEnabled(false);
				}
				return;
			}
		}
	}

	public void valueChanged(TreeSelectionEvent e) {
		DefaultMutableTreeNode selectedNode = (DefaultMutableTreeNode)listTree.getLastSelectedPathComponent();
		if (selectedNode != null && selectedNode instanceof ObservationListTreeNode){
			ObservationListTreeNode listNode = (ObservationListTreeNode)selectedNode;
			if (listNode.getObservationList().getListType() == ObservationList.OBLIST_TYPE_USER){
				removeUserListItem.setEnabled(true);
				removeListMenuItem.setEnabled(true);
				return;
			}
		}
		
		removeUserListItem.setEnabled(false);
		removeListMenuItem.setEnabled(false);
	}

	public void mouseClicked(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();
		
		if (e.getSource() == listTree){
			/* clicked on the list */
			if (e.getClickCount() > 1){
				/* double click */
				TreePath closestPath = listTree.getClosestPathForLocation(x, y);
				Rectangle closestBounds = listTree.getPathBounds(closestPath);
				if (x >= closestBounds.x && x < closestBounds.x + closestBounds.width &&
						y >= closestBounds.y && y < closestBounds.y + closestBounds.height){
					DefaultMutableTreeNode closestNode = (DefaultMutableTreeNode)
							closestPath.getLastPathComponent();
					if (closestNode instanceof ObservationListTreeNode){
						ObservationListTreeNode obNode = (ObservationListTreeNode)closestNode;
						if (!tableSetManager.isObservationVisible(obNode.getObservationList())){
							ObservationPoint point = new ObservationPoint(
								((Double)latSpnr.getValue()).doubleValue(),
								((Double)lonSpnr.getValue()).doubleValue(),
								(Date)yearSpnr.getValue(), timezoneOffset, isDST);
							
							tableSetManager.openTabForList(obNode.getObservationList(), point);
						} else {
							tableSetManager.setTabVisible(obNode.getObservationList());
						}
					}
				}
			}
		}
	}

	public void mousePressed(MouseEvent e) {
	}

	public void mouseReleased(MouseEvent e) {
	}

	public void mouseEntered(MouseEvent e) {
	}

	public void mouseExited(MouseEvent e) {
	}

	public void entitySelected(StellarEntity entity) {
		updateDetailsPane();
	}

	public void mapSelectionChanged() {
		updateMaps();
	}
}
